/*
   kopkop - stealthy secure remote event triggering for ip networks

   Copyright (C) 2004 Erno Rigo <mcree@tricon.hu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#define _GNU_SOURCE 1
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dotconf.h>
#include <syslog.h>
#include <popt.h>
#include "init.h"
#include "twofish.h"
#include "packetengine.h"
#include "config.h"
#include <crypt.h>

config_t config = {
     .configFile = DEFAULT_CONFFILE,
     .pidFile = DEFAULT_PIDFILE,
     .lastIDDir = DEFAULT_LASTIDDIR,
     .sourceIP = "127.0.0.1",
     .sourceInterface = "lo"
};

//static DOTCONF_CB(cb_list);
static DOTCONF_CB(cb_str);
static DOTCONF_CB(cb_ctx_command);
static DOTCONF_CB(cb_ctx_command_end);
static DOTCONF_CB(cb_ctx_command_id);
static DOTCONF_CB(cb_int64);

static int currContext = CTX_ALL;
#define CTX_COMMAND 1

kop_command_t command;

static const configoption_t options[] = {
     {"ListenInterface", ARG_STR,  cb_str,               &(config.listenInterface), CTX_ALL},
     {"PidFile",         ARG_STR,  cb_str,               &(config.pidFile),         CTX_ALL},
     {"Secret",          ARG_STR,  cb_str,               &(config.secret),          CTX_ALL},
     {"SourceIP",        ARG_STR,  cb_str,               &(config.sourceIP),        CTX_ALL},
     {"SourceInterface", ARG_STR,  cb_str,               &(config.sourceInterface), CTX_ALL},
     {"LastIDDir",       ARG_STR,  cb_str,               &(config.lastIDDir),       CTX_ALL},
     {"<Command>",       ARG_NONE, cb_ctx_command,       NULL,                      CTX_ALL},
     {"Id",              ARG_STR,  cb_ctx_command_id,    NULL,                      CTX_COMMAND},
     {"Exec",            ARG_STR,  cb_str,               &(command.exec),           CTX_COMMAND},
     {"Timeout",         ARG_STR,  cb_int64,             &(command.timeout),        CTX_COMMAND},
     {"ExecAfter",       ARG_STR,  cb_str,               &(command.execAfter),      CTX_COMMAND},
     {"</Command>",      ARG_NONE, cb_ctx_command_end,   NULL,                      CTX_COMMAND},
     LAST_OPTION
};

DOTCONF_CB(cb_str)
{
    syslog(LOG_DEBUG, "config(%s:%ld): %s = %s", 
	   cmd->configfile->filename, cmd->configfile->line, cmd->name, cmd->data.str);

    *((char **) (cmd->option->info)) = strdup(cmd->data.str);

    return NULL;
}

DOTCONF_CB(cb_ctx_command)
{
     memset(&command, 0, sizeof(kop_command_t));

     syslog(LOG_DEBUG, "config(%s:%ld): entering %s", 
	    cmd->configfile->filename, cmd->configfile->line, cmd->name);

     currContext=CTX_COMMAND;
     
     return NULL;
}

DOTCONF_CB(cb_ctx_command_end)
{
     kop_command_t *newCommand, *curr;
     
     syslog(LOG_DEBUG, "config(%s:%ld): leaving %s", 
	    cmd->configfile->filename, cmd->configfile->line, cmd->name);

     if(!command.id) {
	  syslog(LOG_ERR, "config(%s:%ld): you must specify a command Id for this command",
		 cmd->configfile->filename, cmd->configfile->line);
     } else if(!command.exec) {
	  syslog(LOG_ERR, "config(%s:%ld): you must specify a command Exec parameter for this command",
		 cmd->configfile->filename, cmd->configfile->line);
     } else {

	  newCommand=malloc(sizeof(kop_command_t));
	  memcpy(newCommand, &command, sizeof(kop_command_t));
	  
	  curr=config.commands;
	  if(!curr) {
	       config.commands=newCommand;
	  } else {
	       while(curr->next) {
		    curr=curr->next;
	       }
	       curr->next=newCommand;    
	  }

     }
     
/*      syslog(LOG_DEBUG, "init: registered command:\nid: %s\nexec: %s\ntimeout: %lli\nexecafter: %s",  */
/* 	    command.id, command.exec, command.timeout, command.execAfter); */

     currContext=CTX_ALL;
     
     return NULL;
}

DOTCONF_CB(cb_ctx_command_id)
{
     syslog(LOG_DEBUG, "config(%s:%ld): <Command> %s = %s", 
	    cmd->configfile->filename, cmd->configfile->line, cmd->name, cmd->data.str);

     command.id=calloc(COMMAND_MAX_LEN,1);

     memcpy(command.id, cmd->data.str,
	    strlen(cmd->data.str)<=COMMAND_MAX_LEN ? strlen(cmd->data.str) : COMMAND_MAX_LEN );
     
     return NULL;
}

DOTCONF_CB(cb_int64)
{
    syslog(LOG_DEBUG, "config(%s:%ld): %s = %s", 
	   cmd->configfile->filename, cmd->configfile->line, cmd->name, cmd->data.str);

    *((u_int64_t *) (cmd->option->info)) = atoll(cmd->data.str);

    return NULL;
}

const char *dotconf_contextchecker(command_t *cmd, unsigned long mask)
{
     if (currContext!=mask) {
	  syslog(LOG_ERR, "config(%s:%ld): error: %s not allowed here",
		 cmd->configfile->filename, cmd->configfile->line, cmd->name);
	  return "yup, that's the way this cruel world goes";
     }
     
     return NULL;
}

FUNC_ERRORHANDLER(dotconf_errorhandler)
{
     syslog(LOG_ERR, "config error: %s", msg);

     /* continue reading the configfile ; return 1 stop after first error found */
     return 0;
}

int readconf()
{
    configfile_t *configfile;

    syslog(LOG_INFO, "reading configuration from %s", config.configFile);

    configfile = dotconf_create(config.configFile, options, NULL, CASE_INSENSITIVE);

    if (!configfile) {
	syslog(LOG_ERR, "error opening config file");
	return 1;
    }

    configfile->errorhandler = (dotconf_errorhandler_t) dotconf_errorhandler;
    configfile->contextchecker = (dotconf_contextchecker_t) dotconf_contextchecker;

    if (dotconf_command_loop(configfile) == 0)
	syslog(LOG_ERR, "error reading config file");

    dotconf_cleanup(configfile);

    return EXIT_SUCCESS;
}

struct poptOption optionsTable[] = {
     {"verbose", 'v', POPT_ARG_NONE, &(config.verbose), 0,
      "Be verbose (print log messages to stderr, do not detach)"},
     {"command", 'c', POPT_ARG_STRING, &(config.command), 0,
      "Command to send to target IP", "STRING"},
     {"listen", 'l', POPT_ARG_NONE, &(config.daemonMode), 0,
      "Start in daemon mode and listen for commands"},
     {"version", 'V', POPT_ARG_NONE, &(config.getVersionInfo), 0,
      "Print \""PACKAGE" "VERSION"\" and exit"},
     {"config", 'f', POPT_ARG_STRING, &(config.configFile), 0,
      "Use different configuration instead of " DEFAULT_CONFFILE, "FILENAME"},
     POPT_AUTOHELP {NULL, 0, 0, NULL, 0}
};


void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error)
	fprintf(stderr, "%s: %s\n", error, addl);
    exit(exitcode);
}

int parseargs(int argc, const char *argv[])
{

    char *targetip;

    char c;			/* used for argument parsing */
    poptContext optCon;		/* context for parsing command-line options */

    optCon = poptGetContext(NULL, argc, argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[options] [target ip]");

    if (argc < 1) {
	poptPrintUsage(optCon, stderr, 0);
	exit(1);
    }

    /* Now do options processing, get portname */
    while ((c = poptGetNextOpt(optCon)) >= 0) {
    }

//     printf("config: %i\n",config.daemonMode);

    targetip = (char *) poptGetArg(optCon);
    if (targetip)
	config.targetIP = strdup(targetip);

    if (c < -1) {
	/* an error occurred during option processing */
	fprintf(stderr, "%s: %s\n", poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
	return 1;
    }


    poptFreeContext(optCon);

    return EXIT_SUCCESS;
}


int init(int argc, const char *argv[])
{
     kop_command_t *command;
     struct crypt_data crypt_r_data;

     parseargs(argc, argv);

     if(getuid()) {
	  fprintf(stderr,"give me power: kopkop requires root privileges\n");
	  exit(EXIT_FAILURE);
     }

     srand(getDevRandomSeed());
     
     if (config.verbose) {
	  setlogmask(LOG_UPTO(LOG_DEBUG));
	  openlog("kopkop", LOG_CONS | LOG_PID | LOG_NDELAY | LOG_PERROR, LOG_AUTHPRIV);
     } else {
	  setlogmask(LOG_UPTO(LOG_INFO));
	  openlog("kopkop", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_AUTHPRIV);
     }
     
     syslog(LOG_INFO, "warming up");
     
     readconf();         

     if (config.error) {
	  syslog(LOG_ERR, "config file errors encountered");
	  return EXIT_FAILURE;
     }

     command=config.commands;
     while(command) {
	  syslog(LOG_DEBUG, "init: registered command: id: %s exec: %s timeout: %lli sec execafter: %s", 
		 command->id, command->exec, command->timeout, command->execAfter);
	  command=command->next;
     }

     config.multihash=strdup(crypt_r(config.secret, MULTIHASH_SALT, &crypt_r_data));

     return EXIT_SUCCESS;
}
