/*
   kopkop - stealthy secure remote event triggering for ip networks

   Copyright (C) 2004 Erno Rigo <mcree@tricon.hu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <libnet.h>

#ifndef init_h_included
#define init_h_included

int init();

#define DEFAULT_CONFFILE "/etc/kopkop/kopkop.conf"
#define DEFAULT_PIDFILE "/var/run/kopkop.pid"
#define DEFAULT_LASTIDDIR "/var/lib/kopkop"

//#define PACKAGE_SALT     "$1$kopkopXX"
#define AUTHTOKEN_SALT   "$1$sesameXX"
#define PACKAGESEAL_SALT "$1$lockitXX"
#define MULTIHASH_SALT   "$1$hasherXX"

#define PACKETID_REPLAY_TRESHOLD 10
#define COMMAND_MAX_LEN 80

typedef struct kop_command_t_t {
     struct kop_command_t_t *next;
     char *id;
     char *exec;
     u_int64_t timeout;
     char *execAfter;
} kop_command_t;

typedef struct {

     char *configFile;                  // configuration file
     char *pidFile;                     // pid file
     char *lastIDDir;                   // last source packet ID files directory

     char *listenInterface;             // interface to listen on
     unsigned int layer2HeaderSize;     // layer2 header size of the listen interface (bytes)
     unsigned int layer2MTU;            // layer2 MTU of the listen or source interfaces (bytes)
     unsigned int randomPadding;        // random packet padding used to spread package sizes up to MTU

     int daemonMode;                    // we are in daemon mode - requested on command line (bool)
     int getVersionInfo;                // we are in 'get version' mode - command line (bool)

     char *secret;                      // shared secret password

     char *sourceInterface;             // source interface
     char *sourceIP;                    // source IP string representation
     u_long sourceIP_bin;               // source IP binary representation
     char *targetIP;                    // target IP string representation
     u_long targetIP_bin;               // target IP binary representation
     
     kop_command_t *commands;           // commands (linked list)

     char *command;                     // command string to send

     int error;                         // errors encountered while parsing config file (bool)
     int verbose;                       // verbose output requested by user (bool)

     char *multihash;                   // password hash for fast multihash

} config_t;

#endif // init_h_included
