/*
   kopkop - stealthy secure remote event triggering for ip networks

   Copyright (C) 2004 Erno Rigo <mcree@tricon.hu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <pcap.h>
#include <libnet.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>
#include <netinet/in.h>
#include <syslog.h>
#include <abz/bcat.h>

#include "init.h"
#include "kopkopd.h"
#include "kopkop.h"
#include "packetengine.h"

#include "glob.inc"

char *strreplace(char *sfrom, char *sto, char *sin) {
     char *sres="";
     char *s=sin;
     char *ss;
     char *bfr=alloca(strlen(sin));

     while((ss=strstr(s, sfrom))) {
	  memset(bfr,0,strlen(sin));
	  strncpy(bfr,s,ss-s);
	  sres=bcat(sres,bfr,sto,NULL);
	  s=&ss[strlen(sfrom)];
     }
     sres=strdup(bcat(sres,s,NULL));
     return sres;
}

void myPCAPHandler(u_char * args, const struct pcap_pkthdr *header, const u_char * packet)
{

     payload_t payload;
     int type=-1;
     kop_command_t *command;
     char sourceIP[16];
     char *exec;
     memset(sourceIP,0,16);

//     if (DEBUG) syslog(LOG_DEBUG, "capture: got packet with length of [%d]", header->caplen);
//     printf("capture: got packet with length of [%d]\n", header->caplen);

     type=getPacket(packet,&payload,sourceIP);
     
     if (type>=0) {
	  syslog(LOG_INFO, "capture: authenticated packet type %i [from:%s] [command:%s]", type, sourceIP, payload.safe.command);
	  command=config.commands;
	  while(command) {
	       if (!strncmp(command->id, payload.safe.command, COMMAND_MAX_LEN)) {
		    syslog(LOG_INFO, "capture: recognized command: id: %s",command->id);
		    syslog(LOG_DEBUG, "capture: exec: %s timeout: %lli sec execafter: %s", 
			   command->exec, command->timeout, command->execAfter);
		    
		    if(!fork()) {
			 // I'M A CHILD - START CRYING
			 exec=strreplace("%sourceip%",sourceIP,command->exec);		    
			 syslog(LOG_INFO,"executing: %s",exec);
			 system(exec);
			 exit(EXIT_SUCCESS);
			 // WHEW
		    }
		    
		    if(command->timeout) {
			 if(!fork()) {
			      // I'M A CHILD - START CRYING
			      sleep(command->timeout);
			      exec=strreplace("%sourceip%",sourceIP,command->execAfter);
			      syslog(LOG_INFO,"timeout: executing: %s",exec);
			      system(exec);			      
			      exit(EXIT_SUCCESS);
			      // WHEW
			 }
		    }

		    

	       }
//	       syslog(LOG_DEBUG, "init: registered command: id: %s exec: %s timeout: %lli sec execafter: %s", 
//		      command->id, command->exec, command->timeout, command->execAfter);
	       command=command->next;
	  }
	  
     }

}

int kopkopd()
{

     pcap_t *handle;		/* Session handle */
     char errbuf[PCAP_ERRBUF_SIZE];	/* Error string */
//     struct bpf_program filter;	/* The compiled filter */
//     char filter_app[] = "";	/* The filter expression */
     bpf_u_int32 mask;		/* Our netmask */
     bpf_u_int32 net;		/* Our IP */
     FILE *fd;
//     libnet_t *l=NULL;
     int soc;
     struct ifreq ifr;

     if (config.pidFile) {
	  syslog(LOG_INFO,"writing pidfile: %s",config.pidFile);
	  fd=fopen(config.pidFile,"w+");
	  if(!fd) { 
	       syslog(LOG_ERR,"cannot open pidfile: %s for writing",config.pidFile);
	       exit (EXIT_FAILURE);
	  }
	  if(!fprintf(fd,"%i\n",getpid())) {
	       syslog(LOG_ERR,"cannot write pidfile: %s",config.pidFile);
	       exit (EXIT_FAILURE);	  
	  }
	  if(fclose(fd)) {
	       syslog(LOG_ERR,"cannot close pidfile: %s",config.pidFile);
	       exit (EXIT_FAILURE);	  
	  }
     }


     // do not wait() for children
     signal(SIGCLD, SIG_IGN);

     syslog(LOG_INFO, "listening on interface %s ...\n", config.listenInterface);

     /* Define the device */
     //dev = pcap_lookupdev(errbuf);
     /* Find the properties for the device */
     pcap_lookupnet(config.listenInterface, &net, &mask, errbuf);
     /* Open the session in promiscuous mode */
     handle = pcap_open_live(config.listenInterface, BUFSIZ*10, 0, 0, errbuf);
     /* Compile and apply the filter */
//    pcap_compile(handle, &filter, filter_app, 0, net);

     syslog(LOG_DEBUG, "phy linktype is: %i\n", pcap_datalink(handle));

/*     struct if_nameindex *ifindex; */
/*     int i; */

/*     ifindex=if_nameindex(); */

/*     i=0; */
/*     while(ifindex[i].if_index) { */
/* 	 syslog(LOG_DEBUG, "found iface: %i %s", ifindex[i].if_index, ifindex[i].if_name); */
/* 	 i++; */
/*     } */

/*     if_freenameindex(ifindex); */

     // determine iface MTU

     if((soc=socket(AF_INET, SOCK_DGRAM, 0))<0) {
	  syslog(LOG_DEBUG, "socket open failed: %s",strerror(errno));
	  exit(EXIT_FAILURE);
     };
    
     strncpy((void *)&(ifr.ifr_name),(void *)(config.listenInterface),IF_NAMESIZE);
    
     if (ioctl(soc,SIOCGIFMTU, &ifr)<0) {
	  syslog(LOG_ERR, "ioctl to get mtu of listen interface %s failed: %s",config.listenInterface,strerror(errno));
	  exit(EXIT_FAILURE);
     }
     syslog(LOG_DEBUG, "physical layer mtu %i", ifr.ifr_mtu);

     // this is just ugly
     switch (pcap_datalink(handle)) {
     case DLT_EN10MB:
     case DLT_IEEE802:
	  config.layer2HeaderSize = 14;
	  break;
     case DLT_SLIP:
     case DLT_SLIP_BSDOS:
	  config.layer2HeaderSize = 16;
	  break;
     case DLT_PPP:
     case DLT_NULL:
#ifdef DLT_PPP_SERIAL
     case DLT_PPP_SERIAL:
#endif
#ifdef DLT_LOOP
     case DLT_LOOP:
#endif
	  config.layer2HeaderSize = 4;
	  break;
     case DLT_PPP_BSDOS:
	  config.layer2HeaderSize = 24;
	  break;
     case DLT_FDDI:
	  config.layer2HeaderSize = 13;
	  break;
     case DLT_RAW:
	  config.layer2HeaderSize = 0;
	  break;
     case DLT_IEEE802_11:
	  config.layer2HeaderSize = 14;
	  break;
#ifdef DLT_TOKEN_RING
     case DLT_TOKEN_RING:
	  config.layer2HeaderSize = 20;
	  break;
#endif
     case DLT_ATM_RFC1483:
#ifdef DLT_CIP
     case DLT_CIP:
#endif
#ifdef DLT_ATM_CLIP
     case DLT_ATM_CLIP:
#endif
	  config.layer2HeaderSize = 8;
	  break;
#ifdef DLT_C_HDLC
     case DLT_C_HDLC:
	  config.layer2HeaderSize = 4;
	  break;
#endif
#ifdef DLT_LINUX_SSL
     case DLT_LINUX_SLL:
#endif
#ifdef DLT_LANE8023
     case DLT_LANE8023:
#endif
	  config.layer2HeaderSize = 16;
	  break;
     default:
	  syslog(LOG_ERR, "physical layer (type: %i) not supported", pcap_datalink(handle));
	  exit(EXIT_FAILURE);
	  break;
     }

//    pcap_setnonblock(handle, 1, errbuf);

     pcap_loop(handle, 0, myPCAPHandler, NULL);

     struct pcap_stat stat;
     pcap_stats(handle,&stat);
     syslog(LOG_DEBUG,"%i packets received %i packets dropped",stat.ps_recv, stat.ps_drop);

     pcap_close(handle);

     return EXIT_SUCCESS;
}
