/*
   kopkop - stealthy secure remote event triggering for ip networks

   Copyright (C) 2004 Erno Rigo <mcree@tricon.hu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#define _GNU_SOURCE 1
#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <pcap.h>
#include <libnet.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/in.h>
#include <syslog.h>
#include <crypt.h>
#include <abz/bcat.h>

#include "init.h"
#include "kopkopd.h"
#include "kopkop.h"
#include "packetengine.h"
#include "twofish.h"
#include "mwcrand.h"

#include "glob.inc"

getPacket_t getICMP;
sendPacket_t sendICMP;
getPacketPayload_t getICMPPayload;

getPacket_t getUDP;
sendPacket_t sendUDP;
getPacketPayload_t getUDPPayload;

getPacket_t getTCP;
sendPacket_t sendTCP;
getPacketPayload_t getTCPPayload;

// update packetengine.h: PACKET_TYPES to reflect your changes here
packetTypes_t packetTypes[] = {
    {0, "ICMP echo-request", sendICMP, getICMP, getICMPPayload},
    {1, "ICMP echo-reply",  sendICMP, getICMP, getICMPPayload},
    {2, "ICMP destination-unreachable",  sendICMP, getICMP, getICMPPayload},
    {3, "UDP",  sendUDP, getUDP, getUDPPayload},
    {4, "TCP",  sendTCP, getTCP, getTCPPayload},
    {-1}
};

#define IP i.ip
#define ICMP i.icmp
#define UDP u.udp
#define TCP t.tcp
#define _(A,B) .mask.A=-1, .pattern.A=B

char *multihash(char *salt, char *seed1, char *seed2)
{
    struct crypt_data data;
    char *hash1, *hash2, *res;
    char *scratch = alloca(22 + 22 + 1);

    data.initialized = 0;

    hash1 = strdupa(crypt_r(seed1, salt, &data));
    hash2 = strdupa(crypt_r(seed2, salt, &data));

    memset(scratch, ' ', 22 + 22);
    memcpy(scratch, &(hash1[12]), strlen(hash1) - 12);
    memcpy(&(scratch[22]), &(hash2[12]), strlen(hash2) - 12);
    scratch[22 + 22] = '\0';

    if (DEBUG) syslog(LOG_DEBUG, "multihash: [hash1:%s] [hash2:%s] [scratch:%s]", hash1, hash2, scratch);

    res = crypt_r(scratch, salt, &data);

    return &(res[12]);
}

char *stringhash64(char *str, u_int64_t salt) {
     int i;
     static char res[23];
     mwc_rand_env_t env;
/*      u_int64_t salt64; */

/*      for(i=0;i<strlen(salt);i++) { */
/* 	  salt64+=i*salt[i]; */
/*      } */

     mwc_rand_seed64(&env, salt);

//     memset(&res,' ',22);

     for(i=0;i<22;i++) 
	  res[i]=(char)mwc_rand(&env);

     for(i=0;i<strlen(str);i++) {
	  res[i%22]^=(char)(str[i] ^ (char)(mwc_rand(&env)));
     }

     res[22]='\0';

     return res;
}

// must be FAST (calculated for every incoming packet)
char *multihash64_fast(u_int64_t salt, char *seed1, char *seed2)
{
    struct crypt_data data;
    char *hash1, *hash2, *res;
    char *scratch = alloca(22 + 22 + 1);

//    return "1234567890123456789012";

    data.initialized = 0;   

    hash1 = strdupa(stringhash64(seed1, salt));
    hash2 = strdupa(stringhash64(seed2, salt));

    memset(scratch, ' ', 22 + 22);
    memcpy(scratch, hash1, 22);
    memcpy(&(scratch[22]), hash2, 22);
    scratch[22 + 22] = '\0';

    if (DEBUG) syslog(LOG_DEBUG, "multihash: [hash1:%s] [hash2:%s] [scratch:%s]", hash1, hash2, scratch);

    res = stringhash64(scratch, salt);

    return &(res[12]);
}

char *multihash64(u_int64_t salt, char *seed1, char *seed2) {
     char saltchars[]="./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
     char saltstring[]="$1$________";
     mwc_rand_env_t env;
     int i;

     mwc_rand_seed64(&env, salt);

     /* Turn it into printable characters from `saltchars'. */
     for (i = 0; i < 8; i++)
	  saltstring[3+i] = saltchars[mwc_rand(&env) % strlen(saltchars)];

     return multihash(saltstring, seed1, seed2);
}

u_int64_t readFile_int64(char *dirname,char *filename) {
     FILE *fd;
     u_int64_t value;
     
     if (DEBUG) syslog(LOG_DEBUG,"readFile64: opening file %s",filename);
/*      if(chdir(dirname)) { */
/* 	  syslog(LOG_ERR,"readFile64: cannot chdir to %s",config.lastIDDir); */
/* 	  exit(EXIT_FAILURE); */
/*      } */
     fd=fopen(bcat(dirname,"/",filename,NULL),"r");
     if(!fd) { 
	  if (DEBUG) syslog(LOG_DEBUG,"readFile64: cannot open %s/%s for reading",dirname,filename);
	  value=0;
     } else {
	  if (!fread(&value,1,sizeof(u_int64_t),fd)) {
	       syslog(LOG_ERR,"readFile64: cannot read %s/%s",dirname,filename);
	       exit(EXIT_FAILURE);
	  }
	  fclose(fd);
     }
     return value;
}

void writeFile_int64(char *dirname, char *filename, u_int64_t value) {
     FILE *fd;
/*      if(chdir(dirname)) { */
/* 	  syslog(LOG_ERR,"writeFile64: cannot chdir to %s",config.lastIDDir); */
/* 	  exit(EXIT_FAILURE); */
/*      } */
     fd=fopen(bcat(dirname,"/",filename,NULL),"w+");
     if(!fd) { 
	  syslog(LOG_ERR,"writeFile64: cannot open %s/%s for writing",dirname,filename);
	  exit (EXIT_FAILURE);
     };
     if(!fwrite(&(value),1,sizeof(u_int64_t),fd)) {
	  syslog(LOG_ERR,"writeFile64: cannot write %s/%s",dirname,filename);
	  exit (EXIT_FAILURE);	  
     }
     if(fclose(fd)) {
	  syslog(LOG_ERR,"writeFile64: cannot close %s/%s",dirname,filename);
	  exit (EXIT_FAILURE);	  
     }
}

int getPacket(const u_char * packet, payload_t *payload, char *sourceIP)
{
     int i;
     packetMaskMask_t *pack_in = (packetMaskMask_t *) (packet + config.layer2HeaderSize);
     char *sip;
     char *key = NULL;
//     payload_t payload;
//     FILE *fd;
     u_int64_t packetid;
     char *filename=NULL;

     if ((sip = libnet_addr2name4(pack_in->i.ip.saddr, LIBNET_DONT_RESOLVE)) == NULL) {
	  syslog(LOG_ERR, "getPacket: cannot resolve source ip address");
	  exit(EXIT_FAILURE);
     }
     memcpy(sourceIP,sip,strlen(sip));

//     key = multihash(PACKAGE_SALT, sourceIP, config.secret);

//     if (DEBUG) syslog(LOG_DEBUG, "getPacket: packet is from %s", sourceIP);
     if (DEBUG) fprintf(stderr, "getPacket: packet is from %s\n", sourceIP);

//     return -1;

     for (i = 0; i < PACKET_TYPES; i++) {
	  packetTypes[i].getPacketPayload(packet, payload);
	  key = strdupa(multihash64_fast(ntohll(payload->salt), config.sourceIP, config.secret));	  
//	  key = strdupa("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa");
	  if (DEBUG) syslog(LOG_DEBUG, "getPacket: key should be %s", key);
	  if (packetTypes[i].getPacket(packet, i, (keySet_t *)key, payload)) {
	       if (DEBUG) syslog(LOG_DEBUG, "getPacket: key was %s", key);

	       syslog(LOG_DEBUG, "getPacket: packet recognized by filter #%i %s", i, packetTypes[i].desc);

	       
	       if (DEBUG) hexdump((char *)payload,sizeof(payload_t) - PAYLOAD_RANDOM_PADDING_SIZE);

	       if (DEBUG) syslog(LOG_DEBUG,"getPacket: payload salt: %llx",ntohll(payload->salt));
	       
	       twofish( (char *)&(payload->safe), 
			sizeof(payload->safe), 
			multihash(PACKAGESEAL_SALT, sourceIP, config.secret), 
			ntohll(payload->salt), 
			TWOFISH_DECODE);

	       if (DEBUG) hexdump((char *)payload,sizeof(payload_t) - PAYLOAD_RANDOM_PADDING_SIZE);

	       if (DEBUG) syslog(LOG_DEBUG,"getPacket: checking package auth token [key:%s]",multihash(AUTHTOKEN_SALT, sourceIP, config.secret));
	       if (memcmp(payload->safe.authToken,
			  multihash(AUTHTOKEN_SALT, sourceIP, config.secret),
			  sizeof(payload->safe.authToken))) {		    
		    syslog(LOG_DEBUG,"getPacket: Package authentication failure from %s",sourceIP);
		    return -1;
	       }

//	       memcpy(filename,sourceIP,strlen(sourceIP));

/* 	       filename=sourceIP; */
/* 	       if (DEBUG) syslog(LOG_DEBUG,"getPacket: opening packetid file %s",filename); */
/* 	       if(!chdir(config.lastIDDir)) { */
/* 		    syslog(LOG_ERR,"getPacket: cannot chdir to %s",config.lastIDDir); */
/* 		    exit(EXIT_FAILURE); */
/* 	       } */
/* 	       fd=fopen(filename,"r"); */
/* 	       if(!fd) {  */
/* 		    if (DEBUG) syslog(LOG_DEBUG,"getPacket: file %s does not exist",filename); */
/* 		    packetid=0; */
/* 	       } else { */
/* 		    if (!fread(&packetid,1,sizeof(u_int64_t),fd)) { */
/* 			 syslog(LOG_ERR,"getPacket: cannot read %s/%s",config.lastIDDir,filename); */
/* 			 exit(EXIT_FAILURE); */
/* 		    } */
/* 		    fclose(fd); */
/* 	       } */

	       packetid=readFile_int64(config.lastIDDir, sourceIP);

	       if (DEBUG) syslog(LOG_DEBUG,"getPacket: checking packet ids [stored:%llx] [got:%llx]",packetid,payload->safe.packetID);
	       if(payload->safe.packetID + PACKETID_REPLAY_TRESHOLD<=packetid) {
		    syslog(LOG_WARNING,
			   "getPacket: Possible replay attack from %s. Command aborted. Remove %s/%s to reset lastID counter.",
			   sourceIP, config.lastIDDir, sourceIP);
		    return -1;
	       }

	       if (DEBUG) syslog(LOG_DEBUG,"getPacket: writing packetid file %s",filename);
	       writeFile_int64(config.lastIDDir, sourceIP, payload->safe.packetID);
/* 	       fd=fopen(filename,"w+"); */
/* 	       fwrite(&(payload->safe.packetID),1,sizeof(u_int64_t),fd); */
/* 	       fclose(fd); */

	       return i;
	  }
     }

     return -1;
}

int sendPacket(int type)
{
     libnet_t *l = NULL;
     char *key = NULL;
     char errbuf[LIBNET_ERRBUF_SIZE];
     payload_t payload;
     u_int64_t salt,seed,packetid;
//     FILE *fd;
     char *filename;
//     u_int32_t mtu=0;
    

     salt=getDevRandomSeed();
     seed=getDevRandomSeed();

     key = strdupa(multihash64_fast(salt, config.sourceIP, config.secret));

     if (DEBUG) syslog(LOG_INFO, "sendPacket: targetting %s with packet type %i [key:%s]", config.targetIP, type, key);

     if ((l = libnet_init(LIBNET_RAW4, config.sourceInterface, errbuf)) == NULL) {
	  syslog(LOG_ERR, "sendPacket: libnet_init() failed: %s", errbuf);
	  exit(EXIT_FAILURE);
     }

     if ((config.sourceIP_bin = libnet_name2addr4(l, config.sourceIP, LIBNET_RESOLVE)) == -1) {
	  syslog(LOG_ERR, "sendPacket: Bad source IP address: %s\n", config.sourceIP);
	  exit(EXIT_FAILURE);
     }

     if ((config.targetIP_bin = libnet_name2addr4(l, config.targetIP, LIBNET_RESOLVE)) == -1) {
	  syslog(LOG_ERR, "sendPacket: Bad target IP address: %s\n", config.targetIP);
	  exit(EXIT_FAILURE);
     }
     
//     memcpy(filename,config.sourceIP,strlen(config.sourceIP));
     filename="lastid";

     packetid=readFile_int64(config.lastIDDir,"lastid");
/*      if (DEBUG) syslog(LOG_DEBUG,"sendPacket: opening packetid file %s",filename); */
/*      chdir(config.lastIDDir); */
/*      fd=fopen(filename,"r"); */
/*      if(!fd) {  */
/* 	  if (DEBUG) syslog(LOG_DEBUG,"sendPacket: file %s does not exist",filename); */
/* 	  packetid=0; */
/*      } else { */
/* 	  fread(&packetid,1,sizeof(u_int64_t),fd); */
/* 	  fclose(fd); */
/*      } */

     packetid++;

     syslog(LOG_DEBUG, "sendPacket: packet id: %llx", packetid);

     if (DEBUG) syslog(LOG_DEBUG,"sendPacket: writing packetid file %s",filename);
     writeFile_int64(config.lastIDDir,"lastid",packetid);
/*      fd=fopen(filename,"w+"); */
/*      fwrite(&packetid,1,sizeof(u_int64_t),fd); */
/*      fclose(fd); */

     memset(&payload,0,sizeof(payload));
     payload.salt=htonll(salt);
     payload.safe.packetID=htonll(packetid);
//     payload.safe.packetSeed=htonll(seed);
     strcpy(&(payload.safe.authToken[0]),multihash(AUTHTOKEN_SALT, config.sourceIP, config.secret));
     strncpy(&(payload.safe.command[0]),config.command,COMMAND_MAX_LEN);

     if (DEBUG) hexdump((char *)&payload,sizeof(payload) - PAYLOAD_RANDOM_PADDING_SIZE);

     twofish( (char *)&payload.safe, 
	      sizeof(payload.safe), 
	      multihash(PACKAGESEAL_SALT, config.sourceIP, config.secret), 
	      salt, 
	      TWOFISH_ENCODE);

     if (DEBUG) syslog(LOG_DEBUG,"sendPacket: payload salt: %llx",salt);

     if (DEBUG) hexdump((char *)&payload,sizeof(payload) - PAYLOAD_RANDOM_PADDING_SIZE);

/*      libnet_ptag_t t; */

/*      t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_ICMPV4_ECHO_H + PAYLOAD_SIZE,	/\* length *\/ */
/* 			   0,	                /\* TOS *\/ */
/* 			   0,	                /\* IP ID *\/ */
/* 			   0,	                /\* IP Frag *\/ */
/* 			   64+(rand()%64),	/\* TTL *\/ */
/* 			   IPPROTO_ICMP,	        /\* protocol *\/ */
/* 			   0,	                /\* checksum *\/ */
/* 			   config.sourceIP_bin,	/\* source IP *\/ */
/* 			   config.targetIP_bin,	/\* destination IP *\/ */
/* 			   NULL,	                /\* payload *\/ */
/* 			   0,	                /\* payload size *\/ */
/* 			   l,	                /\* libnet handle *\/ */
/* 			   0); */
/*      if (t == -1) { */
/* 	  syslog(LOG_ERR, "sendPacket: Can't build IP header: %s\n", libnet_geterror(l)); */
/* 	  exit(EXIT_FAILURE); */
/*      } */

/*      syslog(LOG_DEBUG, "sendPacket: output interface is: %s\n", libnet_getdevice(l));      */

/*      libnet_clear_packet(l); */

     packetTypes[type].sendPacket(l, type, (keySet_t *)key, &payload);

     if (libnet_write(l) == -1) {
	  syslog(LOG_ERR, "sendPacket: Write error: %s\n", libnet_geterror(l));
	  exit(EXIT_FAILURE);
     }

     libnet_destroy(l);

     return EXIT_SUCCESS;
}


int matchPacketMask(const u_char * packet, packetMask_t * packetMask)
{
    packetMask_t *pack_in = (packetMask_t *) (packet + config.layer2HeaderSize);
    char *p = alloca(sizeof(packetMaskMask_t));
    int u;
    char *pmask, *ppattern;

    pmask = (char *) &(packetMask->mask);
    ppattern = (char *) &(packetMask->pattern);

    memcpy(p, pack_in, sizeof(packetMaskMask_t));

    for (u = 0; u < sizeof(packetMaskMask_t); u++) {
	if (DEBUG) if (pmask[u]) syslog(LOG_DEBUG, "matchPacketMask: %.2x %.2x %.2x", (u_int8_t)p[u], (u_int8_t)pmask[u], (u_int8_t)ppattern[u]);
	if ((p[u] & pmask[u]) != ppattern[u])
	    return 0;
    }

    return 1;
}

// --------------------------
// ICMP PACKET HANDLING
// --------------------------

void getICMPPayload(const u_char * packet, payload_t *payload) {
     memcpy(payload,
	    &(packet[ config.layer2HeaderSize + sizeof(struct iphdr) + sizeof(struct icmphdr) ]),
	    PAYLOAD_SIZE - PAYLOAD_RANDOM_PADDING_SIZE);
}

int getICMP(const u_char * packet, int type, keySet_t *key, payload_t *payload)
{
    packetMaskMask_t *pack_in = (packetMaskMask_t *) (packet + config.layer2HeaderSize);

    u_int8_t icmptype=ICMP_ECHO;
    u_int8_t key1=0;

    switch (type) {
    case 0: 
	 icmptype=ICMP_ECHO; 
	 key1 = key->u._8bit.key1;
	 break;
    case 1: 
	 icmptype=ICMP_ECHOREPLY; 
	 key1 = key->u._8bit.key2;
	 break;
    case 2: 
	 icmptype=ICMP_DEST_UNREACH; 
	 key1 = key->u._8bit.key1;
	 break;
    };

    u_int16_t key2 = key->u._16bit.key2;
    u_int16_t key3 = key->u._16bit.key3;

    packetMask_t packetMask = {
	_(IP.protocol, IPPROTO_ICMP),
	_(ICMP.type, icmptype),
	_(ICMP.code, key1),
	_(ICMP.un.echo.id, htons(key2)),
	_(ICMP.un.echo.sequence, htons(key3))
    };

    if (DEBUG) syslog(LOG_DEBUG, "getICMP %i: testing packet for keys %x %x %x", type, key1, key2, key3);

    if (matchPacketMask(packet, &packetMask)) {
	 if (DEBUG) syslog(LOG_DEBUG, 
			   "getICMP %i: got matching packet with keys %x %x %x",
			   type, 
			   pack_in->i.icmp.code, 
			   ntohs(pack_in->i.icmp.un.echo.id), 
			   ntohs(pack_in->i.icmp.un.echo.sequence));

	 getICMPPayload(packet,payload);

	 return 1;
    }

    return 0;
}

void sendICMP(libnet_t * l, int type, keySet_t *key, payload_t *payload)
{
    libnet_ptag_t t;
    u_int8_t icmptype=ICMP_ECHO;
    u_int8_t key1=0;

    switch (type) {
    case 0: 
	 icmptype=ICMP_ECHO; 
	 key1 = key->u._8bit.key1;
	 break;
    case 1: 
	 icmptype=ICMP_ECHOREPLY; 
	 key1 = key->u._8bit.key2;
	 break;
    case 2: 
	 icmptype=ICMP_DEST_UNREACH; 
	 key1 = key->u._8bit.key1;
	 break;
    };

    u_int16_t key2 = key->u._16bit.key2;
    u_int16_t key3 = key->u._16bit.key3;

    if (DEBUG) syslog(LOG_DEBUG, "sendICMP %i: with keys %x %x %x", type, key1, key2, key3);

    t = libnet_build_icmpv4_echo(icmptype,	/* type */
				 key1,	/* code */
				 0,	/* checksum */
				 key2,	/* id */
				 key3,	/* sequence number */
				 (void *)payload,	/* payload */
				 PAYLOAD_SIZE - (rand() % PAYLOAD_RANDOM_PADDING_SIZE),	/* payload size */
				 l,	/* libnet handle */
				 0);
    if (t == -1) {
	syslog(LOG_ERR, "sendICMP: Can't build ICMP header: %s\n", libnet_geterror(l));
	exit(EXIT_FAILURE);
    }

    t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_ICMPV4_ECHO_H + PAYLOAD_SIZE,	/* length */
			  0,	                /* TOS */
			  0,	                /* IP ID */
			  0,	                /* IP Frag */
			  64+(rand()%64),	/* TTL */
			  IPPROTO_ICMP,	        /* protocol */
			  0,	                /* checksum */
			  config.sourceIP_bin,	/* source IP */
			  config.targetIP_bin,	/* destination IP */
			  NULL,	                /* payload */
			  0,	                /* payload size */
			  l,	                /* libnet handle */
			  0);
    if (t == -1) {
	syslog(LOG_ERR, "sendICMP: Can't build IP header: %s\n", libnet_geterror(l));
	exit(EXIT_FAILURE);
    }

}

// --------------------------
// UDP PACKET HANDLING
// --------------------------

void getUDPPayload(const u_char * packet, payload_t *payload) {
     memcpy(payload,
	    &(packet[ config.layer2HeaderSize + sizeof(struct iphdr) + sizeof(struct udphdr) ]),
	    PAYLOAD_SIZE - PAYLOAD_RANDOM_PADDING_SIZE);
}

int getUDP(const u_char * packet, int type, keySet_t *key, payload_t *payload)
{
    packetMaskMask_t *pack_in = (packetMaskMask_t *) (packet + config.layer2HeaderSize);

//    u_int8_t icmptype;
//    u_int8_t key1;

    u_int16_t key1 = key->u._16bit.key1;
    u_int16_t key2 = key->u._16bit.key2;

    packetMask_t packetMask = {
	_(IP.protocol, IPPROTO_UDP),
	_(UDP.source, htons(key1)),
	_(UDP.dest, htons(key2))
    };

//    hexdump(&packetMask.mask,sizeof(packetMask.mask));
//    hexdump(&packetMask.pattern,sizeof(packetMask.pattern));
//    hexdump(pack_in,sizeof(packetMask));

    if (DEBUG) syslog(LOG_DEBUG, "getUDP %i: testing packet for keys %x %x", type, key1, key2);

    if (matchPacketMask(packet, &packetMask)) {
	 if (DEBUG) syslog(LOG_DEBUG, 
			   "getUDP %i: got matching packet with keys %x %x",
			   type, 
			   ntohs(pack_in->u.udp.source), 
			   ntohs(pack_in->u.udp.dest));

	 getUDPPayload(packet,payload);

	 return 1;
    }

    return 0;
}

void sendUDP(libnet_t * l, int type, keySet_t *key, payload_t *payload)
{
    libnet_ptag_t t;
//    u_int8_t icmptype;
    u_int32_t rnd;

    u_int16_t key1 = key->u._16bit.key1;
    u_int16_t key2 = key->u._16bit.key2;

    if (DEBUG) syslog(LOG_DEBUG, "sendUDP %i: with keys %x %x", type, key1, key2);
    
    rnd = rand() % PAYLOAD_RANDOM_PADDING_SIZE;

    t = libnet_build_udp(
	 key1,                                  /* source port */
	 key2,                                  /* destination port */
	 LIBNET_UDP_H + PAYLOAD_SIZE - rnd,     /* packet length */
	 0,                                     /* checksum */
	 (void *)payload,                       /* payload */
	 PAYLOAD_SIZE - rnd,                    /* payload size */
	 l,                                     /* libnet handle */
	 0);                                    /* libnet id */

    if (t == -1) {
	syslog(LOG_ERR, "sendUDP: Can't build header: %s\n", libnet_geterror(l));
	exit(EXIT_FAILURE);
    }

    t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_UDP_H + PAYLOAD_SIZE,	/* length */
			  0,	                /* TOS */
			  0,	                /* IP ID */
			  0,	                /* IP Frag */
			  64+(rand()%64),	/* TTL */
			  IPPROTO_UDP,	        /* protocol */
			  0,	                /* checksum */
			  config.sourceIP_bin,	/* source IP */
			  config.targetIP_bin,	/* destination IP */
			  NULL,	                /* payload */
			  0,	                /* payload size */
			  l,	                /* libnet handle */
			  0);
    if (t == -1) {
	syslog(LOG_ERR, "sendUDP: Can't build IP header: %s\n", libnet_geterror(l));
	exit(EXIT_FAILURE);
    }

}

// --------------------------
// TCP PACKET HANDLING
// --------------------------

void getTCPPayload(const u_char * packet, payload_t *payload) {
     memcpy(payload,
	    &(packet[ config.layer2HeaderSize + sizeof(struct iphdr) + sizeof(struct tcphdr) ]),
	    PAYLOAD_SIZE - PAYLOAD_RANDOM_PADDING_SIZE);
}

int getTCP(const u_char * packet, int type, keySet_t *key, payload_t *payload)
{
     packetMaskMask_t *pack_in = (packetMaskMask_t *) (packet + config.layer2HeaderSize);

//    u_int8_t icmptype;
//    u_int8_t key1;

     u_int16_t key1 = key->u._16bit.key1;
     u_int16_t key2 = key->u._16bit.key2;
     u_int32_t key3 = key->u._32bit.key3;
     u_int32_t key4 = key->u._32bit.key4;

     packetMask_t packetMask = {
	  _(IP.protocol, IPPROTO_TCP),
	  _(TCP.source, htons(key1)),
	  _(TCP.dest, htons(key2)),
	  _(TCP.seq, htonl(key3)),
	  _(TCP.ack_seq, htonl(key4)),
     };

/*     hexdump(&packetMask.mask,sizeof(packetMask.mask)); */
/*     hexdump(&packetMask.pattern,sizeof(packetMask.pattern)); */
/*     hexdump(pack_in,sizeof(packetMask)); */

     if (DEBUG) syslog(LOG_DEBUG, "getTCP %i: testing packet for keys %x %x %lx %lx", type, key1, key2, (unsigned long)key3, (unsigned long)key4);

     if (matchPacketMask(packet, &packetMask)) {
	  if (DEBUG) syslog(LOG_DEBUG, 
			    "getTCP %i: got matching packet with keys %x %x %lx %lx",
			    type, 
			    ntohs(pack_in->t.tcp.source), 
			    ntohs(pack_in->t.tcp.dest), 
			    (unsigned long)ntohl(pack_in->t.tcp.seq), 
			    (unsigned long)ntohl(pack_in->t.tcp.ack_seq));

	  getTCPPayload(packet,payload);

	  return 1;
     }

     return 0;
}

void sendTCP(libnet_t * l, int type, keySet_t *key, payload_t *payload)
{
    libnet_ptag_t t;
//    u_int8_t icmptype;
    u_int32_t rnd;

    u_int16_t key1 = key->u._16bit.key1;
    u_int16_t key2 = key->u._16bit.key2;
    u_int32_t key3 = key->u._32bit.key3;
    u_int32_t key4 = key->u._32bit.key4;

    if (DEBUG) syslog(LOG_DEBUG, "sendTCP %i: with keys %x %x %lx %lx", type, key1, key2, (unsigned long)key3, (unsigned long)key4);
    
    rnd = rand() % PAYLOAD_RANDOM_PADDING_SIZE;

    t = libnet_build_tcp(
	 key1,                                    /* source port */
	 key2,                                    /* destination port */
	 key3,                                    /* sequence number */
	 key4,                                    /* acknowledgement num */
	 rand(),                                  /* control flags */
	 32767+(rand()%32767),                    /* window size */
	 0,                                       /* checksum */
	 rand(),                                  /* urgent pointer */
	 LIBNET_TCP_H + PAYLOAD_SIZE - rnd,       /* TCP packet size */
	 (void *)payload,                         /* payload */
	 PAYLOAD_SIZE - rnd,                      /* payload size */
	 l,                                       /* libnet handle */
	 0);                                      /* libnet id */
    if (t == -1) {
	syslog(LOG_ERR, "sendTCP: Can't build header: %s\n", libnet_geterror(l));
	exit(EXIT_FAILURE);
    }

    t = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_UDP_H + PAYLOAD_SIZE,	/* length */
			  0,	                /* TOS */
			  0,	                /* IP ID */
			  0,	                /* IP Frag */
			  64+(rand()%64),	/* TTL */
			  IPPROTO_TCP,	        /* protocol */
			  0,	                /* checksum */
			  config.sourceIP_bin,	/* source IP */
			  config.targetIP_bin,	/* destination IP */
			  NULL,	                /* payload */
			  0,	                /* payload size */
			  l,	                /* libnet handle */
			  0);
    if (t == -1) {
	syslog(LOG_ERR, "sendTCP: Can't build IP header: %s\n", libnet_geterror(l));
	exit(EXIT_FAILURE);
    }

}

#undef _
#undef IP
#undef ICMP
#undef UDP
#undef TCP
