/*
   kopkop - stealthy secure remote event triggering for ip networks

   Copyright (C) 2004 Erno Rigo <mcree@tricon.hu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#define _GNU_SOURCE

#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/in.h>

#include <byteswap.h>

#include "init.h"

#if __BYTE_ORDER == __BIG_ENDIAN
#  define ntohl(x)       (x)
#  define ntohs(x)       (x)
#  define htonl(x)       (x)
#  define htons(x)       (x)
#  define htonll(x)      (x)
#  define ntohll(x)      (x)
#  define ntoh(x)        (x)
#  define hton(x)        (x)
#else
#  if __BYTE_ORDER == __LITTLE_ENDIAN
#    define ntohl(x)     __bswap_32 (x)
#    define ntohs(x)     __bswap_16 (x)
#    define htonl(x)     __bswap_32 (x)
#    define htons(x)     __bswap_16 (x)
#    define htonll(x)    __bswap_64 (x)
#    define ntohll(x)    __bswap_64 (x)
#    define ntoh(x)      __bswap (x)
#    define hton(x)      __bswap (x)
#  endif
#endif

typedef union { 
     struct { struct iphdr ip; struct icmphdr icmp; char payload; } i;
     struct { struct iphdr ip; struct udphdr udp; char payload; } u;
     struct { struct iphdr ip; struct tcphdr tcp; char payload; } t;
} packetMaskMask_t;

typedef struct {
/*      unsigned int n; */
/*      char *desc; */
/*      void *generate; */
     packetMaskMask_t mask;
     packetMaskMask_t pattern;
} packetMask_t;


// MTU 576 should be enough for everyone 8)
#define PAYLOAD_RANDOM_PADDING_SIZE (576 - sizeof(struct iphdr) - sizeof(struct tcphdr) - COMMAND_MAX_LEN - 50)

typedef struct {     
     u_int64_t salt;
     struct {
	  u_int64_t packetID;
//	  u_int64_t packetSeed;
	  char authToken[22];
	  char command[COMMAND_MAX_LEN];
	  char randomPadding[PAYLOAD_RANDOM_PADDING_SIZE];
     } safe;
} payload_t;

#define PAYLOAD_SIZE sizeof(payload_t)

typedef struct {
     union {
	  struct {
	       u_int8_t key1;
	       u_int8_t key2;
	       u_int8_t key3;
	       u_int8_t key4;
	       u_int8_t key5;
	       u_int8_t key6;
	       u_int8_t key7;
	       u_int8_t key8;
	       u_int8_t key9;
	       u_int8_t key10;
	       u_int8_t key11;
	       u_int8_t key12;
	       u_int8_t key13;
	       u_int8_t key14;
	       u_int8_t key15;
	       u_int8_t key16;
	  } _8bit;
	  struct {
	       u_int16_t key1;
	       u_int16_t key2;
	       u_int16_t key3;
	       u_int16_t key4;
	       u_int16_t key5;
	       u_int16_t key6;
	       u_int16_t key7;
	       u_int16_t key8;
	  } _16bit;
	  struct {
	       u_int32_t key1;
	       u_int32_t key2;
	       u_int32_t key3;
	       u_int32_t key4;
	  } _32bit;
	  struct {
	       u_int64_t key1;
	       u_int64_t key2;
	  } _64bit;
     } u;
} keySet_t;

typedef void (sendPacket_t)(libnet_t *l, int type, keySet_t *key, payload_t *payload);
typedef int (getPacket_t)(const u_char *packet, int type, keySet_t *key, payload_t *payload);
typedef void (getPacketPayload_t)(const u_char *packet, payload_t *payload);

#define PACKET_TYPES 5

typedef struct {
     int id;
     char *desc;
     sendPacket_t *sendPacket;
     getPacket_t *getPacket;
     getPacketPayload_t *getPacketPayload;
} packetTypes_t;

int getPacket(const u_char *packet, payload_t *payload, char *sourceIP);
int sendPacket(int type);

