/*
   kopkop - stealthy secure remote event triggering for ip networks

   Copyright (C) 2004 Erno Rigo <mcree@tricon.hu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <mcrypt.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <mhash.h>
#include "twofish.h"
#include "mwcrand.h"

#define DEBUG 0

void hexdump(unsigned char *buffer, unsigned long buffersize)
{
     char *bfr1 = alloca(80);
     char *bfr2 = alloca(80);
     char *p1, *p2;
     unsigned char c;
     unsigned long i;

     syslog(LOG_DEBUG, "hexdump [size:%li]:", buffersize);

     p1 = bfr1;
     p2 = bfr2;
     for (i = 0; i < ((buffersize / 8) * 8 + (buffersize % 8 ? 8 : 0)); i++) {

	  if (i < buffersize) {
	       sprintf(p1, "%02x ", buffer[i]);
	       c = buffer[i];
	       if (c < 32 || c > 254 || c == 0x90)
		    c = '.';
	       sprintf(p2, "%c", c);
	  } else {
	       sprintf(p1, "   ");
	       sprintf(p2, " ");
	  }
	  p1 += 3;
	  p2++;
	  if (i % 8 == 7) {
	       syslog(LOG_DEBUG, "%04li: %s: %s", i - 7, bfr1, bfr2);
	       p1 = bfr1;
	       p2 = bfr2;
	  }
     }

}

u_int64_t getDevRandomSeed() {
     FILE *fd;
     u_int64_t res;
     int siz;

     fd=fopen("/dev/random","r");
     if (!fd) {
	  syslog(LOG_ERR,"cannot open /dev/random");
	  exit(EXIT_FAILURE);
     }
     siz=fread(&res,sizeof(u_int64_t),1,fd);
     if (!siz) {
	  syslog(LOG_ERR,"cannot read from /dev/random");
	  exit(EXIT_FAILURE);
     }
     fclose(fd);
     return res;
}

int twofish(u_int8_t *buffer, u_int32_t buffersize, char *password, u_int64_t seed, int mode)
{

     MCRYPT td;
     int64_t i, u;

     u_int8_t *IV;

     int keysize = 19;		/* 128 bits */
     u_int8_t *key = alloca(keysize);

     mwc_rand_env_t env;

     memset(key, 0, keysize);

     if (DEBUG) syslog(LOG_DEBUG, "twofish: init");

     mhash_keygen(KEYGEN_MCRYPT, MHASH_MD5, 1, key, keysize, NULL, 0, password, strlen(password));

     if (DEBUG) syslog(LOG_DEBUG, "twofish: [password:%s] [key:%s]", password, key);
     if (DEBUG) hexdump(key, keysize);

     td = mcrypt_module_open("twofish", NULL, "ofb", NULL);
     if (td == MCRYPT_FAILED) {
	  syslog(LOG_ERR, "mcrypt failed %i", mode);
	  exit(EXIT_FAILURE);
     }

     IV = alloca(mcrypt_enc_get_iv_size(td));

     
     if (DEBUG) syslog(LOG_DEBUG,"got seed %llx",seed);
   
     mwc_rand_seed(&env, seed >> 32, seed & 0x0FFFFFFFF);
     for (i = 0; i < mcrypt_enc_get_iv_size(td); i++) {
	  u=mwc_rand(&env);
	  IV[i] = (char)u;
     }

//     memset(IV, 0, mcrypt_enc_get_iv_size(td));

     if (DEBUG) hexdump(IV, mcrypt_enc_get_iv_size(td));   

     i = mcrypt_generic_init(td, key, keysize, IV);
     if (i < 0) {
	  mcrypt_perror(i);
	  return 1;
     }

     if (DEBUG) hexdump(buffer, buffersize);

     for (i = 0; i < buffersize; i++) {
	  switch (mode) {
	  case TWOFISH_ENCODE:
	       u = mcrypt_generic(td, &(buffer[i]), 1);
	       break;
	  case TWOFISH_DECODE:
	       u = mdecrypt_generic(td, &(buffer[i]), 1);
	       break;
	  }
     }
    
     if (DEBUG) hexdump(buffer, buffersize);
    
     if (DEBUG) syslog(LOG_DEBUG, "twofish: deinit...");
    
     mcrypt_generic_deinit(td);

     mcrypt_module_close(td);

     if (DEBUG) syslog(LOG_DEBUG, "twofish: finished");

     return 0;

}
